import json

from distyll.utils import (
    get_transcripts_from_audio_file,
    get_youtube_metadata,
    download_youtube,
    init_dl_dir,
    get_yt_video_id,
    get_audio_from_video,
)
from distyll.config import DL_DIR
from typing import Union, Dict
from pathlib import Path
import logging


def from_youtube(
    yt_url: str, dl_dir: Union[str, Path] = DL_DIR, openai_apikey: str = None
) -> Dict[str, str]:
    """
    Retrieves the transcript of a YouTube video.

    :param yt_url: The URL of the YouTube video.
    :param dl_dir: (Optional) The directory to download the video to.
    :param openai_apikey: (Optional) OpenAI API key.
    :return: A dictionary containing the video title, the YouTube URL, and the transcript texts.
    """
    logging.info(f"Processing {yt_url}, just getting the video title.")
    # Set up download
    dl_dir = init_dl_dir(dl_dir)
    video_id = get_yt_video_id(yt_url)
    transcript_json = video_id + ".json"
    transcript_json_path = Path(dl_dir) / transcript_json
    yt_filename = video_id + ".mp3"
    yt_out_path = Path(dl_dir) / yt_filename

    if not transcript_json_path.exists():
        if yt_out_path.exists():
            logging.info(
                f"Already downloaded {yt_filename}, just getting the video title."
            )
            video_metadata = get_youtube_metadata(youtube_url=yt_url)
        else:
            logging.info(f"Downloading {yt_filename}, just getting the video title.")
            video_metadata = download_youtube(youtube_url=yt_url, path_out=yt_out_path)

        video_title = video_metadata["title"]
        video_date = video_metadata["upload_date"]
        video_uploader = video_metadata["uploader"]
        channel = video_metadata["channel"]

        transcript_texts = get_transcripts_from_audio_file(
            yt_out_path, openai_apikey=openai_apikey
        )
        transcript_data = {
            "title": video_title,
            "date": video_date,
            "yt_url": yt_url,
            "uploader": video_uploader,
            "channel": channel,
            "transcripts": transcript_texts,
        }
        transcript_json_path.write_text(json.dumps(transcript_data))
        return transcript_data
    else:
        logging.info(f"Already downloaded {video_id}")
        return json.loads(transcript_json_path.read_text())


def from_local_video(video_path: Union[str, Path], openai_apikey: str = None):
    audio_path = get_audio_from_video(video_path)
    transcript_texts = get_transcripts_from_audio_file(
        audio_path, openai_apikey=openai_apikey
    )
    return transcript_texts



from bs4 import BeautifulSoup
from typing import Union, List, Dict, Any, Literal
import requests
import logging
from typing import Union
from pathlib import Path
from openai import OpenAI
import yt_dlp
import os


OPENAI_APIKEY = None


def init_dl_dir(dir_path: Union[str, Path]) -> Path:
    """
    Initializes the download directory.

    Args:
        dir_path (Union[str, Path]): The path to the download directory.

    Returns:
        Path: The path to the initialized download directory.
    """
    if type(dir_path) is str:
        dir_path = Path(dir_path)
    dir_path.mkdir(parents=True, exist_ok=True)
    return dir_path


def set_api_key(openai: str = None) -> bool:
    """
    Set the OpenAI API key as an environment variable.

    Args:
        openai (str): The OpenAI API key.

    Returns:
        bool: True if the OpenAI API key is set, False otherwise.
    """
    if openai is not None:
        global OPENAI_APIKEY
        OPENAI_APIKEY = openai
        return True
    return False


def get_openai_client(apikey: Union[str, None] = None) -> OpenAI:
    """
    Helper function to get an OpenAI client
    :param apikey:
    :return:
    """
    global OPENAI_APIKEY
    if apikey is not None:
        oai_client = OpenAI(api_key=apikey)
    elif OPENAI_APIKEY is not None:
        oai_client = OpenAI(api_key=OPENAI_APIKEY)
    else:
        if os.getenv("OPENAI_APIKEY") is not None:
            oai_client = OpenAI(api_key=os.getenv("OPENAI_APIKEY"))
        else:
            raise ValueError("OpenAI API key not provided.")
    return oai_client


def get_arxiv_title(arxiv_url: str) -> Union[str, None]:
    """
    Helper function to get the title of an ArXiV paper
    :param arxiv_url:
    :return:
    """
    logging.info(f"Getting arXiV title from {arxiv_url}")
    response = requests.get(arxiv_url)
    if response.status_code != 200:
        logging.info(
            f"Failed to get the page. HTTP status code: {response.status_code}"
        )
        return None

    soup = BeautifulSoup(response.text, "html.parser")
    title_element = soup.find("meta", {"name": "citation_title"})
    if title_element:
        return title_element["content"]
    else:
        logging.info("Failed to find the title element")
        return None


def chunk_text_by_num_words(
    source_text: str, max_chunk_words: int = 100, overlap_fraction: float = 0.25
) -> List[str]:
    """
    Chunk text input into a list of strings, using a number of words
    :param source_text: Input string to be chunked
    :param max_chunk_words: Maximum length of chunk, in words
    :param overlap_fraction: Overlap as a percentage of chunk_words. The overlap is prepended to each chunk.
    :return: return a list of words
    """
    logging.info(f"Chunking text of {len(source_text)} chars by number of words.")
    sep = " "
    overlap_words = int(max_chunk_words * overlap_fraction)

    source_text = source_text.strip()
    word_list = source_text.split(sep)
    chunks_list = list()

    n_chunks = ((len(word_list) - 1 + overlap_words) // max_chunk_words) + 1
    for i in range(n_chunks):
        window_words = word_list[
            max(max_chunk_words * i - overlap_words, 0) : max_chunk_words * (i + 1)
        ]
        chunks_list.append(sep.join(window_words))
    return chunks_list


def chunk_text_by_num_chars(
    source_text: str, max_chunk_chars: int = 300, overlap_fraction: float = 0.25
) -> List[str]:
    """
    Chunk text input into a list of strings
    :param source_text: Input string to be chunked
    :param max_chunk_chars: Maximum length of chunk, in words
    :param overlap_fraction: Overlap as a percentage of chunk_words
    :return: return a list of words
    """
    overlap_chars = int(max_chunk_chars * overlap_fraction)

    source_text = source_text.strip()
    chunks_list = list()

    n_chunks = ((len(source_text) - 1 + overlap_chars) // max_chunk_chars) + 1
    for i in range(n_chunks):
        chunk = source_text[
            max(max_chunk_chars * i - overlap_chars, 0) : max_chunk_chars * (i + 1)
        ]
        chunks_list.append(chunk)
    return chunks_list


def remove_multiple_whitespaces(source_text: str) -> str:
    """
    Replace multiple whitespaces with single space
    :param source_text:
    :return:
    """
    import re

    source_text = re.sub(r"\s+", " ", source_text)
    return source_text


def chunk_text(
    source_text: str,
    method: Literal["words", "chars"],
    token_length: Union[None, int] = None,
) -> List[str]:
    """
    Chunk longer text
    :param source_text: Input text
    :param method: "words" or "chars"
    :param token_length: Number of tokens to chunk by
    :return:
    """
    logging.info(
        f"Chunking text of {len(source_text)} characters with {method} method."
    )
    source_text = remove_multiple_whitespaces(source_text)
    if method == "words":
        return chunk_text_by_num_words(source_text, max_chunk_words=token_length)
    elif method == "chars":
        return chunk_text_by_num_chars(source_text, max_chunk_chars=token_length)
    else:
        raise ValueError(f"Unsupported method: {method}")


def extract_metadata(video_info: Dict[str, Any]) -> Dict[str, Any]:
    metadata = dict()
    for k in ["title", "upload_date", "channel", "uploader"]:
        if k in video_info:
            metadata[k] = video_info[k]
    return metadata


def download_youtube(youtube_url: str, path_out: Path) -> str:
    """
    Download a YouTube video's audio and return its title
    :param youtube_url: URL of the YouTube video
    :param path_out: Path where the audio file will be saved
    :return: Video title
    """
    path_template = str(path_out.absolute())
    if path_template.endswith(".mp3"):
        path_template = path_template[:-4]

    yt_dlp_params = {
        "extract_audio": True,
        "format": "bestaudio/best",
        "audioformat": "mp3",
        "outtmpl": path_template,
        "quiet": True,
        "cachedir": False,
        "postprocessors": [
            {
                "key": "FFmpegExtractAudio",
                "preferredcodec": "mp3",
                "preferredquality": "192",
            }
        ],
    }

    with yt_dlp.YoutubeDL(yt_dlp_params) as video:
        result = video.extract_info(youtube_url, download=True)
        metadata = extract_metadata(result)
        video_title = result["title"]
        logging.info(f"Found {video_title} - downloading")
        video.download(youtube_url)
        logging.info(f"Successfully downloaded to {path_out}")

    return metadata


def get_youtube_metadata(youtube_url: str) -> Dict[str, str]:
    """
    Download a YouTube video and return its metadata
    :param youtube_url:
    :param path_out:
    :param audio_only:
    :return: Video title
    """
    ydl_opts = {
        "quiet": True,
        "extract_flat": True,
        "force_generic_extractor": True,
    }

    with yt_dlp.YoutubeDL(ydl_opts) as ydl:
        result = ydl.extract_info(youtube_url, download=False)
        metadata = extract_metadata(result)
    return metadata


def download_youtube_video(youtube_url: str, path_out: Path) -> str:
    """
    Download a YouTube video's video
    :param youtube_url: URL of the YouTube video
    :param path_out: Path where the video file will be saved
    :return: Video title
    """
    path_template = str(path_out.absolute())
    if path_template.endswith(".mp4"):
        path_template = path_template[
            :-4
        ]  # Just in case, but usually this should be handled externally

    # Update yt_dlp parameters for video download
    yt_dlp_params = {
        "format": "bestvideo[height<=1080][ext=mp4]+bestaudio[ext=m4a]/best[height<=1080][ext=mp4]",  # Choose best video up to 1080p and best audio
        "outtmpl": f"{path_template}.%(ext)s",  # Define output template
        "quiet": True,
        "cachedir": False,
    }

    with yt_dlp.YoutubeDL(yt_dlp_params) as video:
        result = video.extract_info(youtube_url, download=True)
        video_title = result.get(
            "title", "Downloaded Video"
        )  # Fallback to a default title if not available
        logging.info(f"Found {video_title} - downloading")
        logging.info(f"Successfully downloaded to {path_out}")

    return video_title


def get_audio_from_video(video_path: Union[str, Path]) -> Path:
    from moviepy.editor import VideoFileClip

    if type(video_path) == str:
        video_pathobj = Path(video_path)
    else:
        video_pathobj = Path(video_path)

    audio_path = video_pathobj.with_suffix(".mp3")
    video = VideoFileClip(str(video_pathobj))

    audio = video.audio
    audio.write_audiofile(audio_path)

    # Close the video file
    video.close()

    return audio_path


def get_transcripts_from_audio_file(
    audio_file_path: Path,
    max_segment_len: int = 900,
    openai_apikey: Union[str, None] = None,
) -> List[str]:
    """
    Get transcripts of audio files using
    :param audio_file_path:
    :param max_segment_len:
    :param openai_apikey:
    :return:
    """
    oai_client = get_openai_client(openai_apikey)
    clip_outpaths = split_audio_files(audio_file_path, max_segment_len)
    transcript_texts = list()
    logging.info(f"Getting transcripts from {len(clip_outpaths)} audio files...")
    for i, clip_outpath in enumerate(clip_outpaths):
        logging.info(f"Processing transcript {i+1} of {len(clip_outpaths)}...")
        with clip_outpath.open("rb") as audio_file:
            transcript = oai_client.audio.transcriptions.create(
                model="whisper-1", file=audio_file
            )
            transcript_texts.append(transcript.text)

    # Clean up
    for clip_outpath in clip_outpaths:
        os.remove(clip_outpath)

    return transcript_texts


def split_audio_files(audio_file_path: Path, max_segment_len: int = 900) -> List[Path]:
    """
    Split long audio files
    (e.g. so that they fit within the allowed size for Whisper)
    :param audio_file_path:
    :param max_segment_len:
    :return: A list of file paths
    """
    from pydub import AudioSegment

    audio = AudioSegment.from_file(str(audio_file_path))
    logging.info(f"Splitting {audio_file_path} to chunks of {max_segment_len} seconds.")
    # Split long audio into segments
    clip_outpaths = list()
    if audio.duration_seconds > max_segment_len:
        n_segments = 1 + int(audio.duration_seconds) // max_segment_len
    else:
        n_segments = 1
    logging.info(f"Splitting audio to {n_segments}")
    for i in range(n_segments):
        start = max(0, (i * max_segment_len) - 5) * 1000
        end = ((i + 1) * max_segment_len) * 1000
        clip = audio[start:end]

        clip_outpath = audio_file_path.with_suffix(f".{i}.mp3")
        outfile = clip.export(str(clip_outpath))
        outfile.close()
        clip_outpaths.append(clip_outpath)
    return clip_outpaths


def get_yt_video_id(input_str: str) -> str:
    """
    Get the YouTube video ID (name) from a URL
    :param input_str:
    :return:
    """
    output_str = input_str

    # Remove the https:// if it exists
    if output_str.startswith("https://"):
        output_str = output_str.split("https://")[-1]

    # Remove the prefix
    if "watch?v=" in output_str:
        output_str = output_str.split("watch?v=")[-1]
    else:
        output_str = output_str.split("/")[-1]

    # Remove the suffix
    if "?" in output_str:
        output_str = output_str.split("?")[0]

    # Return the string
    if len(output_str) == 11:
        return output_str
    else:
        raise ValueError(f"Error parsing the video ID from {input_str}")



from weaviate import WeaviateClient
from weaviate.classes.config import Property, DataType, Configure
from weaviate.util import generate_uuid5
import distyll
from distyll.utils import chunk_text
import distyll.config


def prep_db(client: WeaviateClient) -> None:
    """
    Prepare the database for use
    :param client: Weaviate client
    :return: None
    """
    if client.collections.exists("TextChunk"):
        pass
    else:
        client.collections.create(
            "TextChunk",
            properties=[
                Property(name="title", data_type=DataType.TEXT),
                Property(name="url", data_type=DataType.TEXT, skip_vectorization=True),
                Property(name="chunk", data_type=DataType.TEXT),
                Property(name="chunk_no", data_type=DataType.INT),
            ],
            vectorizer_config=Configure.Vectorizer.text2vec_openai(),
            generative_config=Configure.Generative.openai(
                model=distyll.config.load_gen_model()
            ),
        )


def add_yt_to_db(client: WeaviateClient, yt_url) -> int:
    """
    Add a YouTube video to the database
    :param client: Weaviate client
    :param yt_url: YouTube URL
    :return: Number of chunks added
    """
    prep_db(client)
    transcript_data = distyll.transcripts.from_youtube(yt_url)
    chunks_collection = client.collections.get("TextChunk")
    chunk_no = 0
    with chunks_collection.batch.fixed_size() as batch:
        for t in transcript_data["transcripts"]:
            for _, transcript in enumerate(transcript_data["transcripts"]):
                for chunk in chunk_text(transcript):
                    batch.add_object(
                        properties={
                            "title": transcript_data["title"],
                            "url": transcript_data["yt_url"],
                            "chunk": chunk,
                            "chunk_no": chunk_no,
                        },
                        uuid=generate_uuid5(chunk),
                    )
                    chunk_no += 1
    print(f"Added {chunk_no} chunks to the database")
    return chunk_no



from distyll.utils import get_arxiv_title
from distyll.utils import (
    init_dl_dir,
)
from distyll.config import DL_DIR
from pypdf import PdfReader
from typing import Union, Dict
from pathlib import Path
import requests
import logging


def _download_pdf(pdf_url: str, dl_dir: Union[str, Path] = DL_DIR) -> Path:
    """
    Get the text from a PDF and parse it
    :param pdf_url:
    :param dl_dir
    :return:
    """
    logging.info(f"Downloading {pdf_url} text")
    pdf_filename = pdf_url.split("/")[-1]

    # Set up download
    dl_dir = init_dl_dir(dl_dir)
    out_path = Path(dl_dir) / pdf_filename

    # Does file exist already
    if out_path.exists():
        return out_path

    # Get PDF file
    else:
        response = requests.get(pdf_url)
        with out_path.open("wb") as f:
            f.write(response.content)

    return out_path


def _parse_pdf(pdf_path: Union[Path, str]) -> str:
    """
    Read contents of a PDF files
    :param pdf_path:
    :return:
    """
    logging.info(f"Parsing text from {pdf_path}")
    if type(pdf_path) == str:
        pdf_path = Path(pdf_path)
    pdf_reader = PdfReader(pdf_path)

    # Initialize a string to store the text content
    pdf_text = ""
    n_pages = len(pdf_reader.pages)

    # Iterate through the pages and extract the text
    for page_num in range(n_pages):
        page = pdf_reader.pages[page_num]
        pdf_text += "\n" + page.extract_text()
    return pdf_text


def from_pdf(pdf_url: str) -> str:
    """
    Downloads a PDF file from the specified URL and parses its text content.

    :param pdf_url: The URL of the PDF file to download and parse.
    :return: The parsed text content of the PDF file.
    """
    logging.info(f"Downloading and reading text from {pdf_url}")
    pdf_path = _download_pdf(pdf_url)
    pdf_text = _parse_pdf(pdf_path)
    return pdf_text


def from_arxiv_paper(arxiv_url: str) -> Union[Dict[str, str], None]:
    """
    Retrieve arXiv paper information.

    :param arxiv_url: The URL of the arXiv paper.
    :return: A dictionary containing the title, URL, and text of the arXiv paper.
    """
    logging.info(f"Getting arXiV paper from {arxiv_url}")
    if "arxiv.org" not in arxiv_url:
        logging.info("URL is not from arxiv.org")
        return None

    # Get Arxiv paper ID
    arxiv_id = arxiv_url.split("/")[-1]
    if arxiv_id.endswith(".pdf"):
        arxiv_id = arxiv_id[:-4]
    txt_path = Path(DL_DIR) / f"{arxiv_id}.txt"

    # Get title
    title = get_arxiv_title(f"https://arxiv.org/abs/{arxiv_id}")

    # Check if text exists already
    if txt_path.exists():
        pdf_text = txt_path.read_text()
        return {"title": title, "url": arxiv_url, "text": pdf_text}
    else:
        pdf_text = from_pdf(f"https://arxiv.org/pdf/{arxiv_id}.pdf")
        with txt_path.open("w") as f:
            f.write(pdf_text)
        return {"title": title, "url": arxiv_url, "text": pdf_text}
